import pytest
import okta.models as models
from tests.mocks import MockOktaClient


"""
Testing Okta Models Instantiation in different scenarios
"""


@pytest.mark.vcr()
@pytest.mark.asyncio
@ pytest.mark.parametrize("group_obj",
                          [
                              {  # No Okta Model included
                                  'profile': {
                                      'name': 'GroupName',
                                      'description': 'Group Description'
                                  }
                              },
                              models.Group({  # Parent Okta model included
                                  "profile": {
                                      'name': 'GroupName',
                                      'description': 'Group Description'
                                  }
                              }),
                              models.Group({  # Parent + Nested models included
                                  "profile": models.GroupProfile({
                                      "name": "GroupName",
                                      "description": "Group Description"
                                  })
                              })
                          ])
async def test_create_group_different_inputs(fs, group_obj):
    # Instantiate Mock Client
    client = MockOktaClient(fs)

    # Create Group
    group, _, err = await client.create_group(group_obj)
    assert err is None
    assert isinstance(group, models.Group)
    if isinstance(group_obj, dict):
        assert group.profile.name == group_obj["profile"]["name"]
        assert group.profile.description == group_obj["profile"]["description"]
    else:
        assert group.profile.name == group_obj.profile.name
        assert group.profile.description == group_obj.profile.description

    # Delete created group
    _, err = await client.delete_group(group.id)
    assert err is None



from pydash.strings import camel_case


class APIClient():
    """
    A base class for the Okta API Clients.
    """

    def __init__(self):
        pass

    def form_response_body(self, body: dict):
        """
        Method to verify the response body from the Okta API before
        passing it into the constructor.

        Args:
            body (dict): Okta API response body
        """
        result = {}
        for key, val in body.items():
            if val is None:
                continue
            if not isinstance(val, dict):
                result[camel_case(key)] = val
            else:
                result[camel_case(key)] = self.form_response_body(val)
        return result



# flake8: noqa
"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from okta.okta_object import OktaObject


class IdentityProviderCredentialsClient(
    OktaObject
):
    """
    A class for IdentityProviderCredentialsClient objects.
    """

    def __init__(self, config=None):
        super().__init__(config)
        if config:
            self.client_id = config["clientId"]\
                if "clientId" in config else None
            self.client_secret = config["clientSecret"]\
                if "clientSecret" in config else None
        else:
            self.client_id = None
            self.client_secret = None

    def request_format(self):
        parent_req_format = super().request_format()
        current_obj_format = {
            "client_id": self.client_id,
            "client_secret": self.client_secret
        }
        parent_req_format.update(current_obj_format)
        return parent_req_format



import pytest
import re
import os

TEST_OKTA_URL = "test"
B_TEST_OKTA_URL = b"test"
URL_REGEX = r"dev-\d+"
B_URL_REGEX = rb"dev-\d+"
PYTEST_MOCK_CLIENT = "pytest_mock_client"
PYTEST_RE_RECORD = "record_mode"


def pytest_generate_tests(metafunc):
    ''' just to attach the cmd-line args to a test-class that needs them '''
    record_mode = metafunc.config.getoption(PYTEST_RE_RECORD)
    # check if this function is in a test-class that needs the cmd-line args
    if record_mode == "rewrite":
        os.environ[PYTEST_MOCK_CLIENT] = "1"


@pytest.fixture(scope='module')
def vcr_config():
    return {
        # Remove personal details from Integration Tests
        "before_record_request": before_record_request,
        "before_record_response": before_record_response
    }


def before_record_request(request):
    request.uri = re.sub(URL_REGEX, TEST_OKTA_URL, request.uri)
    if "authorization" in request.headers:
        if request.headers["authorization"].startswith("SSWS"):
            request.headers["authorization"] = "SSWS myAPIToken"
        else:
            request.headers["authorization"] = "Bearer myOAuthToken"

    return request


def before_record_response(response):
    response["url"] = re.sub(URL_REGEX, TEST_OKTA_URL, response["url"])
    if "body" in response:
        if "string" in response["body"]:
            # body = response["body"]["string"]
            response["body"]["string"] = re.sub(
                B_URL_REGEX, B_TEST_OKTA_URL, response["body"]["string"])

    # if "Public-Key-Pins-Report-Only" in response["headers"]:
    #     del response["headers"]["Public-Key-Pins-Report-Only"]
    if "content-security-policy-report-only" in response["headers"]:
        response["headers"]["content-security-policy-report-only"] = re.sub(
            URL_REGEX, TEST_OKTA_URL,
            response["headers"]["content-security-policy-report-only"])
    if "link" in response["headers"]:
        current = response["headers"]["link"]
        response["headers"]["link"] = re.sub(
            URL_REGEX, TEST_OKTA_URL, current)

    return response


@pytest.fixture(scope="session", autouse=True)
def cleanup(request):
    """
    To run at the end of a test run
    """
    def clean_up_env_vars():
        if PYTEST_MOCK_CLIENT in os.environ:
            del os.environ[PYTEST_MOCK_CLIENT]
    request.addfinalizer(clean_up_env_vars)



class OktaCollection():
    "Class to build lists composed of OktaObject datatypes"

    @staticmethod
    def form_list(collection: list, data_type: type):
        if not collection:
            # If empty list or None
            return []
        for index in range(len(collection)):
            if not OktaCollection.is_formed(collection[index], data_type):
                collection[index] = data_type(collection[index])
        return collection

    @staticmethod
    def is_formed(value, data_type: type):
        return isinstance(value, data_type)



from urllib.parse import urlencode, quote
from okta.jwt import JWT
from okta.http_client import HTTPClient


class OAuth:
    """
    This class contains the OAuth actions for the Okta Client.
    """
    OAUTH_ENDPOINT = "/oauth2/v1/token"

    def __init__(self, request_executor, config):
        self._request_executor = request_executor
        self._config = config
        self._access_token = None

    def get_JWT(self):
        """
        Generates JWT using client configuration

        Returns:
            str: Generated JSON Web Token
        """
        org_url = self._config["client"]["orgUrl"]
        client_id = self._config["client"]["clientId"]
        private_key = self._config["client"]["privateKey"]

        return JWT.create_token(org_url, client_id, private_key)

    async def get_access_token(self):
        """
        Retrieves or generates the OAuth access token for the Okta Client

        Returns:
            str, Exception: Tuple of the access token, error that was raised
            (if any)
        """
        # Return token if already generated
        if self._access_token:
            return self._access_token

        # Otherwise create new one
        # Get JWT and create parameters for new Oauth token
        jwt = self.get_JWT()
        parameters = {
            'grant_type': 'client_credentials',
            'scope': ' '.join(self._config["client"]["scopes"]),
            'client_assertion_type':
                'urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
            'client_assertion': jwt
        }

        encoded_parameters = urlencode(parameters, quote_via=quote)
        org_url = self._config["client"]["orgUrl"]
        url = f"{org_url}{OAuth.OAUTH_ENDPOINT}?" + \
            encoded_parameters

        # Craft request
        oauth_req, err = await self._request_executor.create_request(
            "POST", url, None, {
                'Accept': "application/json",
                'Content-Type': 'application/x-www-form-urlencoded'
            }, oauth=True)

        # TODO Make max 1 retry
        # Shoot request
        if err:
            return (None, err)
        _, res_details, res_json, err = \
            await self._request_executor.fire_request(oauth_req)
        # Return HTTP Client error if raised
        if err:
            return (None, err)

        # Check response body for error message
        parsed_response, err = HTTPClient.check_response_for_error(
            url, res_details, res_json)
        # Return specific error if found in response
        if err:
            return (None, err)

        # Otherwise set token and return it
        self._access_token = parsed_response["access_token"]
        return (self._access_token, None)

    def clear_access_token(self):
        """
        Clear currently used OAuth access token, probably expired
        """
        self._access_token = None



"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from urllib.parse import urlencode
from okta.http_client import HTTPClient
from okta.models.identity_provider\
    import IdentityProvider
from okta.models.json_web_key\
    import JsonWebKey
from okta.models.csr\
    import Csr
from okta.models.identity_provider_application_user\
    import IdentityProviderApplicationUser
from okta.models.social_auth_token\
    import SocialAuthToken
from okta.utils import format_url
from okta.api_client import APIClient


class IdentityProviderClient(APIClient):
    """
    A Client object for the IdentityProvider resource.
    """

    def __init__(self):
        self._base_url = ""

    async def list_identity_providers(
            self, query_params={}
    ):
        """
        Enumerates IdPs in your organization with pagination. A
        subset of IdPs can be returned that match a supported
        filter expression or query.
        Args:
            query_params {dict}: Map of query parameters for request
            [query_params.q] {str}
            [query_params.after] {str}
            [query_params.limit] {str}
            [query_params.type] {str}
        Returns:
            list: Collection of IdentityProvider instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, IdentityProvider)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(IdentityProvider(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def create_identity_provider(
            self, identity_provider
    ):
        """
        Adds a new IdP to your organization.
        Args:
            {identity_provider}
        Returns:
            IdentityProvider
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps
            """)

        if isinstance(identity_provider, dict):
            body = identity_provider
        else:
            body = identity_provider.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, IdentityProvider)

        if error:
            return (None, response, error)

        try:
            result = IdentityProvider(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_identity_provider_keys(
            self, query_params={}
    ):
        """
        Enumerates IdP key credentials.
        Args:
            query_params {dict}: Map of query parameters for request
            [query_params.after] {str}
            [query_params.limit] {str}
        Returns:
            list: Collection of JsonWebKey instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/credentials/keys
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(JsonWebKey(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def create_identity_provider_key(
            self, json_web_key
    ):
        """
        Adds a new X.509 certificate credential to the IdP key
        store.
        Args:
            {json_web_key}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/credentials/keys
            """)

        if isinstance(json_web_key, dict):
            body = json_web_key
        else:
            body = json_web_key.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def delete_identity_provider_key(
            self, keyId
    ):
        """
        Deletes a specific IdP Key Credential by `kid` if it is
        not currently being used by an Active or Inactive IdP.
        Args:
            key_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/credentials/keys/{keyId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_identity_provider_key(
            self, keyId
    ):
        """
        Gets a specific IdP Key Credential by `kid`
        Args:
            key_id {str}
        Returns:
            JsonWebKey
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/credentials/keys/{keyId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def delete_identity_provider(
            self, idpId
    ):
        """
        Removes an IdP from your organization.
        Args:
            idp_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_identity_provider(
            self, idpId
    ):
        """
        Fetches an IdP by `id`.
        Args:
            idp_id {str}
        Returns:
            IdentityProvider
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, IdentityProvider)

        if error:
            return (None, response, error)

        try:
            result = IdentityProvider(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def update_identity_provider(
            self, idpId, identity_provider
    ):
        """
        Updates the configuration for an IdP.
        Args:
            idp_id {str}
            {identity_provider}
        Returns:
            IdentityProvider
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}
            """)

        if isinstance(identity_provider, dict):
            body = identity_provider
        else:
            body = identity_provider.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, IdentityProvider)

        if error:
            return (None, response, error)

        try:
            result = IdentityProvider(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_csrs_for_identity_provider(
            self, idpId
    ):
        """
        Enumerates Certificate Signing Requests for an IdP
        Args:
            idp_id {str}
        Returns:
            list: Collection of Csr instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/credentials/csrs
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Csr)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(Csr(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def generate_csr_for_identity_provider(
            self, idpId, csr_metadata
    ):
        """
        Generates a new key pair and returns a Certificate Sign
        ing Request for it.
        Args:
            idp_id {str}
            {csr_metadata}
        Returns:
            Csr
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/credentials/csrs
            """)

        if isinstance(csr_metadata, dict):
            body = csr_metadata
        else:
            body = csr_metadata.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Csr)

        if error:
            return (None, response, error)

        try:
            result = Csr(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def revoke_csr_for_identity_provider(
            self, idpId, csrId
    ):
        """
        Revoke a Certificate Signing Request and delete the key
        pair from the IdP
        Args:
            idp_id {str}
            csr_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/credentials/csrs/{csrId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_csr_for_identity_provider(
            self, idpId, csrId
    ):
        """
        Gets a specific Certificate Signing Request model by id
        Args:
            idp_id {str}
            csr_id {str}
        Returns:
            Csr
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/credentials/csrs/{csrId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Csr)

        if error:
            return (None, response, error)

        try:
            result = Csr(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def publish_cer_cert_for_identity_provider(
            self, idpId, csrId, string
    ):
        """
        Update the Certificate Signing Request with a signed X.
        509 certificate and add it into the signing key credent
        ials for the IdP.
        Args:
            idp_id {str}
            csr_id {str}
            {string}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/credentials/csrs/{csrId}
                /lifecycle/publish
            """)

        if isinstance(string, dict):
            body = string
        else:
            body = string.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/x-x509-ca-cert"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def publish_binary_cer_cert_for_identity_provider(
            self, idpId, csrId, string
    ):
        """
        Update the Certificate Signing Request with a signed X.
        509 certificate and add it into the signing key credent
        ials for the IdP.
        Args:
            idp_id {str}
            csr_id {str}
            {string}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/credentials/csrs/{csrId}
                /lifecycle/publish
            """)

        body = HTTPClient.format_binary_data(string)
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/x-x509-ca-cert"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def publish_der_cert_for_identity_provider(
            self, idpId, csrId, string
    ):
        """
        Update the Certificate Signing Request with a signed X.
        509 certificate and add it into the signing key credent
        ials for the IdP.
        Args:
            idp_id {str}
            csr_id {str}
            {string}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/credentials/csrs/{csrId}
                /lifecycle/publish
            """)

        if isinstance(string, dict):
            body = string
        else:
            body = string.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/pkix-cert"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def publish_binary_der_cert_for_identity_provider(
            self, idpId, csrId, string
    ):
        """
        Update the Certificate Signing Request with a signed X.
        509 certificate and add it into the signing key credent
        ials for the IdP.
        Args:
            idp_id {str}
            csr_id {str}
            {string}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/credentials/csrs/{csrId}
                /lifecycle/publish
            """)

        body = HTTPClient.format_binary_data(string)
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/pkix-cert"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def publish_binary_pem_cert_for_identity_provider(
            self, idpId, csrId, string
    ):
        """
        Update the Certificate Signing Request with a signed X.
        509 certificate and add it into the signing key credent
        ials for the IdP.
        Args:
            idp_id {str}
            csr_id {str}
            {string}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/credentials/csrs/{csrId}
                /lifecycle/publish
            """)

        body = HTTPClient.format_binary_data(string)
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/x-pem-file"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_identity_provider_signing_keys(
            self, idpId
    ):
        """
        Enumerates signing key credentials for an IdP
        Args:
            idp_id {str}
        Returns:
            list: Collection of JsonWebKey instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/credentials/keys
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(JsonWebKey(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def generate_identity_provider_signing_key(
            self, idpId, query_params={}
    ):
        """
        Generates a new X.509 certificate for an IdP signing ke
        y credential to be used for signing assertions sent to
        the IdP
        Args:
            idp_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.validityYears] {str}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/credentials/keys/generate
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def get_identity_provider_signing_key(
            self, idpId, keyId
    ):
        """
        Gets a specific IdP Key Credential by `kid`
        Args:
            idp_id {str}
            key_id {str}
        Returns:
            JsonWebKey
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/credentials/keys/{keyId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def clone_identity_provider_key(
            self, idpId, keyId, query_params={}
    ):
        """
        Clones a X.509 certificate for an IdP signing key crede
        ntial from a source IdP to target IdP
        Args:
            idp_id {str}
            key_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.targetIdpId] {str}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/credentials/keys/{keyId}/clone
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def activate_identity_provider(
            self, idpId
    ):
        """
        Activates an inactive IdP.
        Args:
            idp_id {str}
        Returns:
            IdentityProvider
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/lifecycle/activate
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, IdentityProvider)

        if error:
            return (None, response, error)

        try:
            result = IdentityProvider(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def deactivate_identity_provider(
            self, idpId
    ):
        """
        Deactivates an active IdP.
        Args:
            idp_id {str}
        Returns:
            IdentityProvider
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/lifecycle/deactivate
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, IdentityProvider)

        if error:
            return (None, response, error)

        try:
            result = IdentityProvider(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_identity_provider_application_users(
            self, idpId
    ):
        """
        Find all the users linked to an identity provider
        Args:
            idp_id {str}
        Returns:
            list: Collection of IdentityProviderApplicationUser instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/users
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, IdentityProviderApplicationUser)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(IdentityProviderApplicationUser(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def unlink_user_from_identity_provider(
            self, idpId, userId
    ):
        """
        Removes the link between the Okta user and the IdP user
        .
        Args:
            idp_id {str}
            user_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/users/{userId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_identity_provider_application_user(
            self, idpId, userId
    ):
        """
        Fetches a linked IdP user by ID
        Args:
            idp_id {str}
            user_id {str}
        Returns:
            IdentityProviderApplicationUser
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/users/{userId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, IdentityProviderApplicationUser)

        if error:
            return (None, response, error)

        try:
            result = IdentityProviderApplicationUser(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def link_user_to_identity_provider(
            self, idpId, userId, user_identity_provider_link_request
    ):
        """
        Links an Okta user to an existing Social Identity Provi
        der. This does not support the SAML2 Identity Provider
        Type
        Args:
            idp_id {str}
            user_id {str}
            {user_identity_provider_link_request}
        Returns:
            IdentityProviderApplicationUser
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/users/{userId}
            """)

        if isinstance(user_identity_provider_link_request, dict):
            body = user_identity_provider_link_request
        else:
            body = user_identity_provider_link_request.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, IdentityProviderApplicationUser)

        if error:
            return (None, response, error)

        try:
            result = IdentityProviderApplicationUser(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_social_auth_tokens(
            self, idpId, userId
    ):
        """
        Fetches the tokens minted by the Social Authentication
        Provider when the user authenticates with Okta via Soci
        al Auth.
        Args:
            idp_id {str}
            user_id {str}
        Returns:
            list: Collection of SocialAuthToken instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/idps/{idpId}/users/{userId}/credentials
                /tokens
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, SocialAuthToken)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(SocialAuthToken(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)



from okta.cache.cache import Cache
import time


class OktaCache(Cache):
    """
    This is a base class implementing a Cache using TTL and TTI.
    Implementing the okta.cache.cache.Cache abstract class.
    """

    def __init__(self, ttl, tti):
        """
        Constructor.

        Arguments:
            ttl {float} -- Time to Live: for cache entries
            tti {float} -- Time to Idle: for cache entries
        """
        super()  # Inherit from parent class
        self._store = {}  # key -> {value, TTI, TTL}
        self._time_to_live = ttl
        self._time_to_idle = tti

    def get(self, key):
        """
        Retrieves value from cache using key

        Arguments:
            key {str} -- Desired key

        Returns:
            str -- Corresponding value to given key
            None -- Unable to find value for this key
        """
        # Get current time
        now = self._get_current_time()
        # Check if key is in cache and valid
        if self.contains(key):
            entry = self._store[key]
            # Reset TTI
            entry["tti"] = now + self._time_to_idle
            # Return desired value and update cache
            self._clean_cache()
            return entry["value"]

        # Return None if key isn't in cache and update cache
        self._clean_cache()
        return None

    def contains(self, key):
        """
        Returns existence of key in cache, as boolean

        Arguments:
            key {str} -- Desired key

        Returns:
            bool -- Existence of key in cache
        """
        return key in self._store and self._is_valid_entry(self._store[key])

    def add(self, key: str, value: tuple):
        """
        Adds a key-value pair to the cache.

        Arguments:
            key {str} -- Key in pair
            value {tuple} -- Tuple of response and response body
        """
        if type(key) == str and\
                (type(value) != list or type(value[1]) != list):
            # Get current time
            now = self._get_current_time()

            # Add new entry to cache with timers
            self._store[key] = {
                'value': value,
                'tti': now + self._time_to_idle,
                'ttl': now + self._time_to_live
            }
        # Update cache
        self._clean_cache()

    def delete(self, key):
        """
        Delete a key-value pair from the cache.

        Arguments:
            key {str} -- Desired key
        """
        # Make sure key is in cache
        if key in self._store:
            # Delete entry
            del self._store[key]

    def clear(self):
        """
        Clear the cache.
        """
        self._store.clear()

    def _clean_cache(self):
        """
        Updates cache by removing expired entries at time of call
        """
        expired = []
        # Check every entry
        for key in self._store.keys():
            # If not valid, delete
            if not self._is_valid_entry(self._store[key]):
                expired.append(key)
        # Delete keys
        for expired_key in expired:
            self.delete(expired_key)

    def _is_valid_entry(self, entry):
        """
        Determines if a given cache entry is not expired.

        Args:
            entry (dict): An entry from the cache composed of value,
            TTI, and TTL

        Returns:
            bool: Boolean value representing if entry is expired
        """
        # Get Current time
        now = self._get_current_time()
        # Check timers and compare against current time
        timers = [entry["tti"], entry["ttl"]]
        return not any(timer <= now for timer in timers)

    def _get_current_time(self):
        """
        Helper function to get current time

        Returns:
            float: value representing the number of seconds since the epoch
        """
        return time.time()



import tests.mocks as mocks
import pytest
import time
import datetime as dt
from http import HTTPStatus
from okta.client import Client
from okta.errors.http_error import HTTPError
from okta.utils import convert_date_time_to_seconds
from okta.http_client import HTTPClient
from okta.request_executor import RequestExecutor
from okta.error_messages import ERROR_MESSAGE_429_MISSING_DATE_X_RESET

REQUEST_TIMEOUT = mocks.REQUEST_TIMEOUT
ORG_URL = mocks.ORG_URL
API_TOKEN = mocks.API_TOKEN
CLIENT_ID = mocks.CLIENT_ID
SCOPES = mocks.SCOPES
PRIVATE_KEY = mocks.PRIVATE_KEY
GET_USERS_CALL = mocks.GET_USERS_CALL

CLIENT_CONFIG = {'orgUrl': ORG_URL, 'token': API_TOKEN}


@pytest.mark.asyncio
async def test_max_retries_no_timeout(monkeypatch, mocker):
    client = Client(user_config=CLIENT_CONFIG)
    query_params = {"limit": "1"}

    monkeypatch.setattr(HTTPClient, 'send_request',
                        mocks.mock_GET_HTTP_Client_response_429)

    monkeypatch.setattr(time, 'sleep', mocks.mock_pause_function)

    http_spy = mocker.spy(HTTPClient, 'send_request')

    users, resp, error = await client.list_users(query_params)

    http_spy.assert_called()
    assert http_spy.call_count ==\
        client.get_request_executor()._max_retries + 1
    assert client.get_request_executor()._request_timeout == 0
    assert isinstance(error, HTTPError)
    assert error is not None
    assert error.status == HTTPStatus.TOO_MANY_REQUESTS
    assert resp.get_status() == HTTPStatus.TOO_MANY_REQUESTS


@pytest.mark.asyncio
async def test_backoff_calculation():
    client = Client(user_config=CLIENT_CONFIG)

    response_429 = (await mocks.mock_GET_HTTP_Client_response_429())[1]
    # ^ has a 1 second difference in retry and datetime
    # backoff should be 2 by Okta standards

    retry_limit_reset = float(response_429.headers["X-Rate-Limit-Reset"])
    date_time = convert_date_time_to_seconds(response_429.headers["Date"])

    backoff_time = client.get_request_executor(
    ).calculate_backoff(retry_limit_reset, date_time)

    assert(backoff_time == 2)


@pytest.mark.asyncio
async def test_no_x_date_header(monkeypatch):
    client = Client(user_config=CLIENT_CONFIG)

    monkeypatch.setattr(HTTPClient, 'send_request',
                        mocks.mock_GET_HTTP_Client_response_429_no_x_reset)
    monkeypatch.setattr(time, 'sleep', mocks.mock_pause_function)

    users, resp, error = await client.list_users()

    assert error is not None
    assert isinstance(error, Exception)
    assert error.args[0] == ERROR_MESSAGE_429_MISSING_DATE_X_RESET
    assert users is None
    assert resp is None


@pytest.mark.asyncio
async def test_no_x_reset_header(monkeypatch):
    client = Client(user_config=CLIENT_CONFIG)

    monkeypatch.setattr(HTTPClient, 'send_request',
                        mocks.mock_GET_HTTP_Client_response_429_no_x_reset)
    monkeypatch.setattr(time, 'sleep', mocks.mock_pause_function)

    users, resp, error = await client.list_users()

    assert error is not None
    assert isinstance(error, Exception)
    assert error.args[0] == ERROR_MESSAGE_429_MISSING_DATE_X_RESET
    assert users is None
    assert resp is None


@pytest.mark.asyncio
async def test_multiple_x_reset_headers(monkeypatch, mocker):
    client = Client(user_config=CLIENT_CONFIG)

    backoff_spy = mocker.spy(RequestExecutor, 'calculate_backoff')

    monkeypatch.setattr(HTTPClient, 'send_request',
                        mocks.mock_GET_HTTP_Client_response_429_multi_x_reset)
    # the X Rate Limit Resets used are 1 sec and 2 sec after the Date header,
    # -> the min. one should be used (1 sec. after) and the backoff calculated
    #    should be equal to 2 (by Okta standards)
    monkeypatch.setattr(time, 'sleep', mocks.mock_pause_function)

    users, resp, error = await client.list_users()

    assert error is not None
    assert users is None
    assert backoff_spy.spy_return == 2


@pytest.mark.asyncio
async def test_req_timeout(monkeypatch):
    this_config = CLIENT_CONFIG.copy()
    this_config.update({"requestTimeout": 1})
    client = Client(user_config=this_config)

    request = await mocks.mock_GET_HTTP_request()

    monkeypatch.setattr(HTTPClient, 'send_request',
                        mocks.mock_GET_HTTP_Client_response_429)

    now = dt.datetime.now(dt.timezone.utc)
    now = now.replace(microsecond=0)
    two_sec_before = (now - dt.timedelta(seconds=2)).timestamp()

    a, b, c, error = await client.get_request_executor(
    ).fire_request_helper(request, 0, two_sec_before)

    assert isinstance(error, Exception)
    assert "Request Timeout exceeded." == error.args[0]



"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from urllib.parse import urlencode
from okta.models.user_factor\
    import UserFactor
from okta.models.security_question\
    import SecurityQuestion
from okta.models.verify_user_factor_response\
    import VerifyUserFactorResponse
from okta.utils import format_url
from okta.api_client import APIClient
from okta.constants import find_factor_model


class UserFactorClient(APIClient):
    """
    A Client object for the UserFactor resource.
    """

    def __init__(self):
        self._base_url = ""

    async def list_factors(
            self, userId
    ):
        """
        Enumerates all the enrolled factors for the specified u
        ser
        Args:
            user_id {str}
        Returns:
            list: Collection of UserFactor instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/factors
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, UserFactor)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(
                    find_factor_model(item["factorType"])(
                        self.form_response_body(item)
                        )
                    )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def enroll_factor(
            self, userId, user_factor, query_params={}
    ):
        """
        Enrolls a user with a supported factor.
        Args:
            user_id {str}
            {user_factor}
            query_params {dict}: Map of query parameters for request
            [query_params.updatePhone] {str}
            [query_params.templateId] {str}
            [query_params.tokenLifetimeSeconds] {str}
            [query_params.activate] {str}
        Returns:
            UserFactor
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/factors
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        if isinstance(user_factor, dict):
            body = user_factor
        else:
            body = user_factor.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, UserFactor)

        if error:
            return (None, response, error)

        try:
            body = self.form_response_body(response.get_body())
            result = find_factor_model(body["factorType"])(body)
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_supported_factors(
            self, userId
    ):
        """
        Enumerates all the supported factors that can be enroll
        ed for the specified user
        Args:
            user_id {str}
        Returns:
            list: Collection of UserFactor instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/factors/catalog
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, UserFactor)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(
                    find_factor_model(item["factorType"])(
                        self.form_response_body(item)
                        )
                    )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_supported_security_questions(
            self, userId
    ):
        """
        Enumerates all available security questions for a user'
        s `question` factor
        Args:
            user_id {str}
        Returns:
            list: Collection of SecurityQuestion instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/factors/questions
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, SecurityQuestion)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(SecurityQuestion(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def delete_factor(
            self, userId, factorId
    ):
        """
        Unenrolls an existing factor for the specified user, al
        lowing the user to enroll a new factor.
        Args:
            user_id {str}
            factor_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/factors/{factorId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_factor(
            self, userId, factorId
    ):
        """
        Fetches a factor for the specified user
        Args:
            user_id {str}
            factor_id {str}
        Returns:
            UserFactor
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/factors/{factorId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, UserFactor)

        if error:
            return (None, response, error)

        try:
            body = self.form_response_body(response.get_body())
            result = find_factor_model(body["factorType"])(body)
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def activate_factor(
            self, userId, factorId, activate_factor_request
    ):
        """
        The `sms` and `token:software:totp` factor types requir
        e activation to complete the enrollment process.
        Args:
            user_id {str}
            factor_id {str}
            {activate_factor_request}
        Returns:
            UserFactor
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/factors/{factorId}/lifecycle
                /activate
            """)

        if isinstance(activate_factor_request, dict):
            body = activate_factor_request
        else:
            body = activate_factor_request.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, UserFactor)

        if error:
            return (None, response, error)

        try:
            body = self.form_response_body(response.get_body())
            result = find_factor_model(body["factorType"])(body)
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def get_factor_transaction_status(
            self, userId, factorId, transactionId
    ):
        """
        Polls factors verification transaction for status.
        Args:
            user_id {str}
            factor_id {str}
            transaction_id {str}
        Returns:
            VerifyUserFactorResponse
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/factors/{factorId}
                /transactions/{transactionId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, VerifyUserFactorResponse)

        if error:
            return (None, response, error)

        try:
            result = VerifyUserFactorResponse(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def verify_factor(
            self, userId, factorId, verify_factor_request, query_params={}
    ):
        """
        Verifies an OTP for a `token` or `token:hardware` facto
        r
        Args:
            user_id {str}
            factor_id {str}
            {verify_factor_request}
            query_params {dict}: Map of query parameters for request
            [query_params.templateId] {str}
            [query_params.tokenLifetimeSeconds] {str}
        Returns:
            VerifyUserFactorResponse
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/factors/{factorId}/verify
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        if isinstance(verify_factor_request, dict):
            body = verify_factor_request
        else:
            body = verify_factor_request.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, VerifyUserFactorResponse)

        if error:
            return (None, response, error)

        try:
            result = VerifyUserFactorResponse(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)



from okta.constants import FINDING_OKTA_DOMAIN, REPO_URL
from okta.error_messages import ERROR_MESSAGE_ORG_URL_MISSING, \
    ERROR_MESSAGE_API_TOKEN_DEFAULT, ERROR_MESSAGE_API_TOKEN_MISSING, \
    ERROR_MESSAGE_AUTH_MODE_INVALID, ERROR_MESSAGE_CLIENT_ID_MISSING, \
    ERROR_MESSAGE_CLIENT_ID_DEFAULT, ERROR_MESSAGE_SCOPES_PK_MISSING, \
    ERROR_MESSAGE_ORG_URL_NOT_HTTPS, ERROR_MESSAGE_ORG_URL_YOUROKTADOMAIN, \
    ERROR_MESSAGE_ORG_URL_TYPO, ERROR_MESSAGE_ORG_URL_ADMIN, \
    ERROR_MESSAGE_PROXY_MISSING_HOST, ERROR_MESSAGE_PROXY_MISSING_AUTH, \
    ERROR_MESSAGE_PROXY_INVALID_PORT


class ConfigValidator():
    """
    This class performs validation checks on the Okta Client configuration.
    """

    def __init__(self, config):
        self._config = config
        self.validate_config()

    """
    Configuration Validators
    """

    def validate_config(self):
        """
        This method validates the client configuration and validates
        the values provided. Throws a ValueError if anything is invalid.

        Raises:
            ValueError: A configuration provided needs to be corrected.
        """
        errors = []
        client = self._config.get('client')
        # check org url
        errors += \
            self._validate_org_url(
                client.get('orgUrl', ""))
        # check proxy settings if provided
        if "proxy" in client:
            errors += self._validate_proxy_settings(client["proxy"])
        # check API details based on authorization mode
        if client.get('authorizationMode') == "SSWS":
            errors += \
                self._validate_token(
                    client.get('token', ""))
        elif client.get('authorizationMode') == "PrivateKey":
            client_fields = ['clientId', 'scopes', 'privateKey']
            client_fields_values = [self._config.get(
                'client').get(field, "") for field in client_fields]
            errors += self._validate_client_fields(*client_fields_values)
        else:  # Not a valid authorization mode
            errors += [
                (f"{ERROR_MESSAGE_AUTH_MODE_INVALID}"
                 f"{client.get('authorizationMode', None)}")]

        if errors:  # raise exception and stop program if errors
            newline = '\n'
            raise ValueError(f"{newline}Errors:"
                             f"{newline + newline.join(errors) + 2*newline}"
                             f"See {REPO_URL} for usage")

    def _validate_client_fields(self, client_id, client_scopes,
                                client_private_key):
        client_fields_errors = []

        # check client id
        client_id = client_id.strip().lower()
        # null or empty string
        if not client_id:
            client_fields_errors.append(ERROR_MESSAGE_CLIENT_ID_MISSING)
        # contains {clientId}
        if "{clientId}".lower() in client_id:
            client_fields_errors.append(ERROR_MESSAGE_CLIENT_ID_DEFAULT)

        # check that at least 1 scope is provided and private key is provided
        if not (client_scopes and client_private_key):
            client_fields_errors.append(ERROR_MESSAGE_SCOPES_PK_MISSING)

        return client_fields_errors

    def _validate_token(self, token: str):
        # remove whitespaces and lowercase token for comparisons
        token = token.strip().lower()

        token_errors = []

        # token is null or empty
        if not token:
            token_errors.append(ERROR_MESSAGE_API_TOKEN_MISSING)
        # token contains {apiToken}
        if "{apiToken}".lower() in token:
            token_errors.append(ERROR_MESSAGE_API_TOKEN_DEFAULT)

        return token_errors

    def _validate_org_url(self, url: str):
        # remove whitespaces and lowercase URL for comparisons
        url = url.strip().lower()

        url_errors = []

        # if empty or null string
        if not url:
            url_errors.append(ERROR_MESSAGE_ORG_URL_MISSING)
        # if url is not https (in non-testing env)
        if url and not (self._config["testing"]["testingDisableHttpsCheck"]
                        or url.startswith('https')):
            url_errors.append(
                (f"{ERROR_MESSAGE_ORG_URL_NOT_HTTPS} Current value: "
                 f"{url or None}. "
                 "You can copy your domain from the Okta Developer Console. "
                 "Follow these instructions to find it: "
                 f"{FINDING_OKTA_DOMAIN}"))
        # url still contains default {yourOktaDomain}
        if "{yourOktaDomain}".lower() in url\
                or "{{yourOktaDomain}}".lower() in url:
            url_errors.append(ERROR_MESSAGE_ORG_URL_YOUROKTADOMAIN)
        # url contains -admin string
        admin_strings = ["-admin.okta.com",
                         "-admin.oktapreview.com", "-admin.okta-emea.com"]
        if any(string in url for string in admin_strings) or "-admin" in url:
            url_errors.append(
                (f"{ERROR_MESSAGE_ORG_URL_ADMIN} "
                 f"Current value: {url}. You can copy your domain "
                 "from the Okta Developer Console. Follow these instructions "
                 f"to find it: {FINDING_OKTA_DOMAIN}"))
        # url ends with .com.com
        if url.endswith(".com.com"):
            url_errors.append((
                f"{ERROR_MESSAGE_ORG_URL_TYPO} Current "
                f"value: {url}. You can copy your domain from "
                "the Okta Developer Console. Follow these instructions to find"
                f" it: {FINDING_OKTA_DOMAIN}"))
        # url contains multiple '://' segments
        if url.count("://") > 1:
            url_errors.append((
                f"{ERROR_MESSAGE_ORG_URL_TYPO} Current "
                f"value: {url}. You can copy your domain from "
                "the Okta Developer Console. Follow these instructions to find"
                f" it: {FINDING_OKTA_DOMAIN}"))

        return url_errors

    def _validate_proxy_settings(self, proxy):
        proxy_errors = []
        if "host" not in proxy:
            proxy_errors.append(ERROR_MESSAGE_PROXY_MISSING_HOST)
        if "username" in proxy and "password" not in proxy or\
                "username" not in proxy and "password" in proxy:
            proxy_errors.append(ERROR_MESSAGE_PROXY_MISSING_AUTH)
        if "port" in proxy:
            try:
                port_number = int(proxy["port"])
                if not 1 <= port_number <= 65535:
                    raise ValueError
            except (TypeError, ValueError):
                proxy_errors.append(ERROR_MESSAGE_PROXY_INVALID_PORT)

        return proxy_errors



# flake8: noqa
"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from okta.okta_object import OktaObject
from okta.okta_collection import OktaCollection
from okta.models import application_accessibility\
    as application_accessibility
from okta.models import application_credentials\
    as application_credentials
from okta.models import application_licensing\
    as application_licensing
from okta.models import application_settings\
    as application_settings
from okta.models import application_sign_on_mode\
    as application_sign_on_mode
from okta.models import application_visibility\
    as application_visibility


class Application(
    OktaObject
):
    """
    A class for Application objects.
    """

    def __init__(self, config=None):
        super().__init__(config)
        if config:
            self.embedded = config["embedded"]\
                if "embedded" in config else None
            self.links = config["links"]\
                if "links" in config else None
            if "accessibility" in config:
                if isinstance(config["accessibility"],
                              application_accessibility.ApplicationAccessibility):
                    self.accessibility = config["accessibility"]
                elif config["accessibility"] is not None:
                    self.accessibility = application_accessibility.ApplicationAccessibility(
                        config["accessibility"]
                    )
                else:
                    self.accessibility = None
            else:
                self.accessibility = None
            self.created = config["created"]\
                if "created" in config else None
            if "credentials" in config:
                if isinstance(config["credentials"],
                              application_credentials.ApplicationCredentials):
                    self.credentials = config["credentials"]
                elif config["credentials"] is not None:
                    self.credentials = application_credentials.ApplicationCredentials(
                        config["credentials"]
                    )
                else:
                    self.credentials = None
            else:
                self.credentials = None
            self.features = OktaCollection.form_list(
                config["features"] if "features"\
                    in config else [],
                str
            )
            self.id = config["id"]\
                if "id" in config else None
            self.label = config["label"]\
                if "label" in config else None
            self.last_updated = config["lastUpdated"]\
                if "lastUpdated" in config else None
            if "licensing" in config:
                if isinstance(config["licensing"],
                              application_licensing.ApplicationLicensing):
                    self.licensing = config["licensing"]
                elif config["licensing"] is not None:
                    self.licensing = application_licensing.ApplicationLicensing(
                        config["licensing"]
                    )
                else:
                    self.licensing = None
            else:
                self.licensing = None
            self.name = config["name"]\
                if "name" in config else None
            self.profile = config["profile"]\
                if "profile" in config else None
            if "settings" in config:
                if isinstance(config["settings"],
                              application_settings.ApplicationSettings):
                    self.settings = config["settings"]
                elif config["settings"] is not None:
                    self.settings = application_settings.ApplicationSettings(
                        config["settings"]
                    )
                else:
                    self.settings = None
            else:
                self.settings = None
            if "signOnMode" in config:
                if isinstance(config["signOnMode"],
                              application_sign_on_mode.ApplicationSignOnMode):
                    self.sign_on_mode = config["signOnMode"]
                elif config["signOnMode"] is not None:
                    self.sign_on_mode = application_sign_on_mode.ApplicationSignOnMode(
                        config["signOnMode"].upper()
                    )
                else:
                    self.sign_on_mode = None
            else:
                self.sign_on_mode = None
            self.status = config["status"]\
                if "status" in config else None
            if "visibility" in config:
                if isinstance(config["visibility"],
                              application_visibility.ApplicationVisibility):
                    self.visibility = config["visibility"]
                elif config["visibility"] is not None:
                    self.visibility = application_visibility.ApplicationVisibility(
                        config["visibility"]
                    )
                else:
                    self.visibility = None
            else:
                self.visibility = None
        else:
            self.embedded = None
            self.links = None
            self.accessibility = None
            self.created = None
            self.credentials = None
            self.features = []
            self.id = None
            self.label = None
            self.last_updated = None
            self.licensing = None
            self.name = None
            self.profile = None
            self.settings = None
            self.sign_on_mode = None
            self.status = None
            self.visibility = None

    def request_format(self):
        parent_req_format = super().request_format()
        current_obj_format = {
            "_embedded": self.embedded,
            "_links": self.links,
            "accessibility": self.accessibility,
            "created": self.created,
            "credentials": self.credentials,
            "features": self.features,
            "id": self.id,
            "label": self.label,
            "lastUpdated": self.last_updated,
            "licensing": self.licensing,
            "name": self.name,
            "profile": self.profile,
            "settings": self.settings,
            "signOnMode": self.sign_on_mode,
            "status": self.status,
            "visibility": self.visibility
        }
        parent_req_format.update(current_obj_format)
        return parent_req_format



from okta.errors.error import Error


class HTTPError(Error):
    def __init__(self, url, response_details, response_body):
        self.status = response_details.status
        self.url = url
        self.response_headers = response_details.headers
        self.stack = ""
        self.message = f"HTTP {self.status} {response_body}"



import os
import okta.models as models
from okta.models.application_sign_on_mode import ApplicationSignOnMode as ASM
from okta.models.factor_type import FactorType as FT
from okta.models.policy_type import PolicyType as PT

DEV_OKTA = "https://developer.okta.com"

FINDING_OKTA_DOMAIN = (f"{DEV_OKTA}"
                       "/docs/guides/find-your-domain/overview")
GET_OKTA_API_TOKEN = (f"{DEV_OKTA}"
                      "/docs/guides/create-an-api-token/overview")
FINDING_OKTA_APP_CRED = (f"{DEV_OKTA}"
                         "/docs/guides/find-your-app-credentials/overview")
REPO_URL = "https://github.com/okta/okta-sdk-python"

EPOCH_YEAR = 1970
EPOCH_MONTH = 1
EPOCH_DAY = 1

DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

_GLOBAL_YAML_PATH = os.path.join(os.path.expanduser('~'), ".okta",
                                 "okta.yaml")
_LOCAL_YAML_PATH = os.path.join(os.getcwd(), "okta.yaml")

OKTA_APP_SIGN_ON_TO_MODEL = {
    ASM.AUTO_LOGIN: models.AutoLoginApplication,
    ASM.BASIC_AUTH: models.BasicAuthApplication,
    ASM.BOOKMARK: models.BookmarkApplication,
    ASM.OPENID_CONNECT: models.OpenIdConnectApplication,
    ASM.SAML_1_1: models.SamlApplication,
    ASM.SAML_2_0: models.SamlApplication,
    ASM.SECURE_PASSWORD_STORE: models.SecurePasswordStoreApplication,
    ASM.WS_FEDERATION: models.WsFederationApplication
}

OKTA_APP_NAME_TO_MODEL = {
    "template_swa": models.SwaApplication,
    "template_swa3field": models.SwaThreeFieldApplication
}

SWA_APP_NAME = "template_swa"
SWA3_APP_NAME = "template_swa3field"


def find_app_model(sign_on_mode, template_name):
    # If ASM found in model map, return model
    if sign_on_mode in OKTA_APP_SIGN_ON_TO_MODEL:
        return OKTA_APP_SIGN_ON_TO_MODEL[sign_on_mode]
    # O/W must be BROWSER PLUGIN APP
    return OKTA_APP_NAME_TO_MODEL[template_name]


OKTA_FACTOR_TYPE_TO_FACTOR = {
    FT.CALL: models.CallUserFactor,
    FT.EMAIL: models.EmailUserFactor,
    FT.PUSH: models.PushUserFactor,
    FT.QUESTION: models.SecurityQuestionUserFactor,
    FT.SMS: models.SmsUserFactor,
    FT.TOKEN: models.TokenUserFactor,
    FT.TOKEN_HARDWARE: models.HardwareUserFactor,
    FT.TOKEN_HOTP: models.CustomHotpUserFactor,
    FT.TOKEN_SOFTWARE_TOTP: models.TotpUserFactor,
    FT.U_2_F: models.U2FUserFactor,
    FT.WEB: models.WebUserFactor,
    FT.WEBAUTHN: models.WebAuthnUserFactor
}


def find_factor_model(factor_type):
    return OKTA_FACTOR_TYPE_TO_FACTOR[factor_type]


OKTA_POLICY_TYPE_TO_MODEL = {
    PT.IDP_DISCOVERY: models.IdentityProviderPolicy,
    PT.OAUTH_AUTHORIZATION_POLICY: models.OAuthAuthorizationPolicy,
    PT.OKTA_SIGN_ON: models.OktaSignOnPolicy,
    PT.PASSWORD: models.PasswordPolicy
}


def find_policy_model(policy_type):
    return OKTA_POLICY_TYPE_TO_MODEL[policy_type]


OKTA_POLICY_RULE_TYPE_TO_MODEL = {
    "PASSWORD": models.PasswordPolicyRule,
    "SIGN_ON": models.OktaSignOnPolicyRule
}


def find_policy_rule_model(policy_rule):
    if policy_rule in OKTA_POLICY_RULE_TYPE_TO_MODEL:
        return OKTA_POLICY_RULE_TYPE_TO_MODEL[policy_rule]
    return models.PolicyRule



from okta.user_agent import UserAgent
from okta.constants import DATETIME_FORMAT
from okta.client import Client
import aiohttp
import asyncio
import json
from yarl import URL
import datetime as dt
import multidict
from tests.conftest import PYTEST_MOCK_CLIENT
import os
import time

REQUEST_TIMEOUT = 5  # seconds
ORG_URL = "https://test.okta.com"
API_TOKEN = "myApiToken"
CLIENT_ID = "yourClientId"
SCOPES = ["okta.scope.1", "okta.scope.2"]
PRIVATE_KEY = "yourPrivateKey"
GET_USERS_CALL = "/api/v1/users"
CLIENT_CONFIG = {'orgUrl': ORG_URL, 'token': API_TOKEN}

# Cache Test Details
TTI = 5.0
TTL = 5.0
CACHE_KEY = ("https://example.com/sample/cache-key/test+test@test."
             "com?with=a&query=string")
CACHE_VALUE = "54321"
ALT_CACHE_KEY = ("https://sample.com/example/cache-key/test+2@test."
                 "com?with=a&query=string")
ALT_CACHE_VALUE = "dbca"


class MockOktaClient(Client):
    def __init__(self, fs):
        if PYTEST_MOCK_CLIENT in os.environ:
            fs.pause()
            super().__init__()
            fs.resume()
        else:
            super().__init__(CLIENT_CONFIG)


async def mock_GET_HTTP_request(*args, **kwargs):
    return {
        "headers": {
            "User-Agent": UserAgent().get_user_agent_string(),
            "Authorization": f"SSWS {API_TOKEN}",
            # etc
        },
        "url": ORG_URL + GET_USERS_CALL,
        "data": {},
        "method": "GET"
    }


class MockHTTPResponseDetails():
    def __init__(self):
        self.status = 200
        self.headers = {
            "Date": "Tue, 1 Jan 2009 00:00:00 GMT",
            "Content-Type": "application/json",
            # etc
        }
        self.url = ORG_URL + GET_USERS_CALL
        self.method = "GET"
        self.links = {}
        self.content_type = "application/json"


class MockHTTP429ResponseDetails():
    def __init__(self):
        now = dt.datetime.now(dt.timezone.utc)
        now = now.replace(microsecond=0)

        one_sec_after = now + dt.timedelta(seconds=1)
        self.status = 429
        self.headers = multidict.CIMultiDict({
            "Date": now.strftime(DATETIME_FORMAT),
            "Content-Type": "application/json",
            "X-Okta-Now": "",
            "X-Rate-Limit-Reset": one_sec_after.timestamp(),
            "X-Okta-Request-id": "okta-request-id",
        })
        self.url = ORG_URL + GET_USERS_CALL
        self.method = "GET"
        self.links = {}
        self.content_type = "application/json"


class MockHTTP429NoXResetResponseDetails():
    def __init__(self):
        now = dt.datetime.now(dt.timezone.utc)
        now = now.replace(microsecond=0)

        self.status = 429
        self.headers = multidict.CIMultiDict({
            "Date": now.strftime(DATETIME_FORMAT),
            "Content-Type": "application/json",
            "X-Okta-Now": "",
            "X-Okta-Request-id": "okta-request-id",
        })
        self.url = ORG_URL + GET_USERS_CALL
        self.method = "GET"
        self.links = {}
        self.content_type = "application/json"


class MockHTTP429NoDateResponseDetails():
    def __init__(self):
        now = dt.datetime.now(dt.timezone.utc)
        now = now.replace(microsecond=0)

        one_sec_after = now + dt.timedelta(seconds=1)
        self.status = 429
        self.headers = multidict.CIMultiDict({
            "Content-Type": "application/json",
            "X-Okta-Now": "",
            "X-Rate-Limit-Reset": one_sec_after.timestamp(),
            "X-Okta-Request-id": "okta-request-id",
        })
        self.url = ORG_URL + GET_USERS_CALL
        self.method = "GET"
        self.links = {}
        self.content_type = "application/json"


class MockHTTP429MultiXResetResponseDetails():
    def __init__(self):
        now = dt.datetime.now(dt.timezone.utc)
        now = now.replace(microsecond=0)

        one_sec_after = now + dt.timedelta(seconds=1)
        two_sec_after = one_sec_after + dt.timedelta(seconds=1)

        base_headers = {
            "Date": now.strftime(DATETIME_FORMAT),
            "Content-Type": "application/json",
            "X-Okta-Now": "",
            "X-Rate-Limit-Reset": one_sec_after.timestamp(),
            "X-Okta-Request-id": "okta-request-id",
        }

        base_headers = list(zip(base_headers.keys(), base_headers.values()))
        base_headers.append(("X-Rate-Limit-Reset", two_sec_after.timestamp()))

        self.status = 429
        self.headers = multidict.CIMultiDict(base_headers)
        self.url = ORG_URL + GET_USERS_CALL
        self.method = "GET"
        self.links = {}
        self.content_type = "application/json"


async def mock_GET_HTTP_Client_response_valid(*args, **kwargs):
    request = await mock_GET_HTTP_request()
    response_details = MockHTTPResponseDetails()
    response_body = '[{"ID": "user.id.1"}]'
    error = None
    return (request, response_details, response_body, error)


async def mock_GET_HTTP_Client_response_valid_with_next(*args, **kwargs):
    request = await mock_GET_HTTP_request()
    response_details = MockHTTPResponseDetails()
    response_details.links = {
        "next": {"url": URL("https://www.next.okta.com")}
    }
    response_body = '[{"ID": "user.id.1"}, {"ID": "user.id.2"}]'
    error = None
    return (request, response_details, response_body, error)


async def mock_GET_HTTP_Client_response_429(*args, **kwargs):
    request = await mock_GET_HTTP_request()
    response_details = MockHTTP429ResponseDetails()
    response_body = '{}'
    error = None
    return (request, response_details, response_body, error)


async def mock_GET_HTTP_Client_response_429_no_x_reset(*args, **kwargs):
    request = await mock_GET_HTTP_request()
    response_details = MockHTTP429NoXResetResponseDetails()
    response_body = '{}'
    error = None
    return (request, response_details, response_body, error)


async def mock_GET_HTTP_Client_response_429_no_date(*args, **kwargs):
    request = await mock_GET_HTTP_request()
    response_details = MockHTTP429NoDateResponseDetails()
    response_body = '{}'
    error = None
    return (request, response_details, response_body, error)


async def mock_GET_HTTP_Client_response_429_multi_x_reset(*args, **kwargs):
    request = await mock_GET_HTTP_request()
    response_details = MockHTTP429MultiXResetResponseDetails()
    response_body = '{}'
    error = None
    return (request, response_details, response_body, error)


async def mock_GET_HTTP_Client_response_error(*args, **kwargs):
    return (None, None, None, json.dumps({
        "errorCode": "",
        "errorSummary": "",
        "errorLink": "",
        "errorId": "",
        "errorCauses": []
    }))


async def mock_timeout_response(*args, **kwargs):
    return (None, None, None, asyncio.TimeoutError())


async def mock_invalid_HTTP_response(*args, **kwargs):
    return (None, None, None, aiohttp.ContentTypeError(None, None))


async def mock_access_token(*args, **kwargs):
    return ("myOAuthToken", None)


def mock_pause_function(*args, **kwargs):
    if PYTEST_MOCK_CLIENT in os.environ:
        time.sleep(args[0])


def mock_cache_return_none(*args, **kwargs):
    return None


def mock_cache_return_value(*args, **kwargs):
    return CACHE_VALUE


SAMPLE_RSA = '''-----BEGIN RSA PRIVATE KEY-----
                MIIEogIBAAKCAQEAvlhONz/qR7dBung7VW
                MhA6V61sGF+VOAELUq7HOr53YxHyyH0Fuh
                ET78wTbRM51dbKVMiY2e3jTTQ8HZyEs6pn
                2apud3gRWa3mmW9sCU9Iu9AaQPVYc9SrfY
                RY6iCvxNVMyfvYrWsbsKxKjLviRL2o3iO2
                gdlZMJfJl+PIm0nJAZ6wdnjNi7uXlVMu48
                xj5Uw8ZBWHGPtDH9ZyYm9HhxYY4CxY2oTV
                6azqnDYUIpFio2sZBcyU+3bnlxcbHI0A4z
                wIiuJbH8KNFcHp13gs5MvsLDFTEx8pMry3
                2rHkCTRrRhad/OsvIVa3J6yvWlL5sz3iai
                SqEjnwBC4nhNv/JKhQIDAQABAoIBAApGZo
                x9dwZhilsIR/2rQPo7Kdcjcbm8i+569SL3
                +IjhXLP32aoB9D8j1Q2SCbd9UHB/uNK5Ei
                EhHWFsOrcg9nzLzXgiiRUvByHn9cYCrdfI
                nk6THl/J0eFxbKuOkElDBuKjvCDFeKL1+5
                YTnZH5UB3vaE24KFatbgbrm0Cl11nXJrXd
                3EUOtjpkPvUo0WzGvK0Ha509ucsUfBCQcy
                mDY6I3Z5bfSNNKsdQrvoHlOONA0nN0uSZR
                IdFgxtvdB8D8yseqg0BN4Bl+/EMA7lprXf
                a+C7vljMCMY9JNtSjuBCM43o3Yy/H6NtzB
                Qa7H/nuzEt5O6s20va12SDA/55PRbVkCgY
                EA28pOgvfOROPVamvz+/kbEEfi1YJx74mv
                20EMVRab1DPkzQnVHVRX53Wqjtd5afLj3W
                2drGQOgLhq2XTXuSh9Ikat1R80NG69klO3
                9X4Uk5+Qk+6c8B84K2w/vCucKT0D8VHO3C
                SdXO3yLIVCJWJjH3NeIK5A6BPs4P71El2Z
                6i0CgYEA3bQjKk8tJCvOxDuo90rbDaZy1U
                Y53/ewmQAgrYZ7ASzvQK0YPvNYIr289nDy
                f1DiVECWsNOoymuNlzMdD2sIP3J9KZ02hz
                5CVbS9W2yIDgf8L3sBZ66fDLIhhAXAUrhM
                k21pC66ccYFGBbnroh+ivKsOeP2lVDjkoB
                mYxbTyULkCgYAaNKbjyrQXVqrtg22Vz3/A
                kzGij2kSTHJhTAIyav2tcXqIT/HPC21ntY
                neeiaJ0WIF1roEAfIQuuxSfTHza7ZvzcmQ
                LaZF0iZcOgsnBB9j8pSEbuDbaH2D/+Yhg3
                mdGR9NOJnXea4hlGVZlZHrby2uZE0GzIAE
                q8uEkCGcTPi7CQKBgEB0N2j8fcYvSjx4j8
                XdKFQOmQHyGa9IRjok9FseZrJPyvrFl+/t
                NSlcegw3h/iFnlcaM/USg9bdb9uOS5uI5W
                v0r/iiMREeg5CBLC/hHKjfn2x6WU04UhsR
                ynakaWZjGDggLdLsn6TMDXxqsdt/UQOCw4
                FfVGpo4+a7qlOIKI5ZAoGAFkC1F269uXAi
                ACUTA1XmTgQ4AiXztKZe1ALUqa4WrgZ6/J
                6XR+VwM4QuWjnlczpF4kHsfHpFi4d4qS3J
                5kMvVUyIY3DwOsQenLo8C+orgmo4UstEjN
                Q1OQt04mV1fp1aE2b9JVjv72cPEE1SMqaE
                p7W0dBcm49L0mY7sNyEFS0U=
                -----END RSA PRIVATE KEY-----'''

SAMPLE_JWK = {'alg': 'RS256',
              'kty': 'RSA',
              'n': '''vlhONz_qR7dBung7VWMhA6V61sGF-
                 VOAELUq7HOr53YxHyyH0FuhET78wTbRM51dbK
                 VMiY2e3jTTQ8HZyEs6pn2apud3gRWa3mmW9sC
                 U9Iu9AaQPVYc9SrfYRY6iCvxNVMyfvYrWsbsK
                 xKjLviRL2o3iO2gdlZMJfJl-PIm0nJAZ6wdnj
                 Ni7uXlVMu48xj5Uw8ZBWHGPtDH9ZyYm9HhxYY
                 4CxY2oTV6azqnDYUIpFio2sZBcyU-3bnlxcbH
                 I0A4zwIiuJbH8KNFcHp13gs5MvsLDFTEx8pMr
                 y32rHkCTRrRhad_OsvIVa3J6yvWlL5sz3iaiS
                 qEjnwBC4nhNv_JKhQ''',
              'e': '''AQAB''',
              'd': '''CkZmjH13BmGKWwhH_atA-jsp1yNxu
                 byL7nr1Ivf4iOFcs_fZqgH0PyPVDZIJt31QcH
                 -40rkSISEdYWw6tyD2fMvNeCKJFS8HIef1xgK
                 t18ieTpMeX8nR4XFsq46QSUMG4qO8IMV4ovX7
                 lhOdkflQHe9oTbgoVq1uBuubQKXXWdcmtd3cR
                 Q62OmQ-9SjRbMa8rQdrnT25yxR8EJBzKYNjoj
                 dnlt9I00qx1Cu-geU440DSc3S5JlEh0WDG290
                 HwPzKx6qDQE3gGX78QwDuWmtd9r4Lu-WMwIxj
                 0k21KO4EIzjejdjL8fo23MFBrsf-e7MS3k7qz
                 bS9rXZIMD_nk9FtWQ''',
              'p': '''28pOgvfOROPVamvz-_kbEEfi1YJx7
                 4mv20EMVRab1DPkzQnVHVRX53Wqjtd5afLj3W
                 2drGQOgLhq2XTXuSh9Ikat1R80NG69klO39X4
                 Uk5-Qk-6c8B84K2w_vCucKT0D8VHO3CSdXO3y
                 LIVCJWJjH3NeIK5A6BPs4P71El2Z6i0''',
              'q': '''3bQjKk8tJCvOxDuo90rbDaZy1UY53
                 _ewmQAgrYZ7ASzvQK0YPvNYIr289nDyf1DiVE
                 CWsNOoymuNlzMdD2sIP3J9KZ02hz5CVbS9W2y
                 IDgf8L3sBZ66fDLIhhAXAUrhMk21pC66ccYFG
                 Bbnroh-ivKsOeP2lVDjkoBmYxbTyULk''',
              'dp': '''GjSm48q0F1aq7YNtlc9_wJMxoo9p
                 EkxyYUwCMmr9rXF6iE_xzwttZ7WJ3nomidFiB
                 da6BAHyELrsUn0x82u2b83JkC2mRdImXDoLJw
                 QfY_KUhG7g22h9g__mIYN5nRkfTTiZ13muIZR
                 lWZWR628trmRNBsyABKvLhJAhnEz4uwk''',
              'dq': '''QHQ3aPx9xi9KPHiPxd0oVA6ZAfIZ
                 r0hGOiT0Wx5msk_K-sWX7-01KVx6DDeH-IWeV
                 xoz9RKD1t1v245Lm4jla_Sv-KIxER6DkIEsL-
                 EcqN-fbHpZTThSGxHKdqRpZmMYOCAt0uyfpMw
                 NfGqx239RA4LDgV9Uamjj5ruqU4gojlk''',
              'qi': '''FkC1F269uXAiACUTA1XmTgQ4AiXz
                 tKZe1ALUqa4WrgZ6_J6XR-VwM4QuWjnlczpF4
                 kHsfHpFi4d4qS3J5kMvVUyIY3DwOsQenLo8C-
                 orgmo4UstEjNQ1OQt04mV1fp1aE2b9JVjv72c
                 PEE1SMqaEp7W0dBcm49L0mY7sNyEFS0U'''}



from okta.jwt import JWT
import tests.mocks as mocks
import os
import pytest

"""
Testing Private Key Inputs
"""


@ pytest.mark.parametrize("jwk_input",
                          [mocks.SAMPLE_JWK, str(mocks.SAMPLE_JWK)])
def test_private_key_PEM_JWK_dict(jwk_input):
    generated_pem, generated_jwk = JWT.get_PEM_JWK(jwk_input)

    assert generated_pem is not None and generated_jwk is not None
    assert not generated_jwk.is_public()


def test_private_key_PEM_JWK_file(fs):
    file_path = os.path.join(os.path.dirname(
        __file__), "test.pem")
    fs.create_file(file_path, contents=mocks.SAMPLE_RSA)
    generated_pem, generated_jwk = JWT.get_PEM_JWK(file_path)

    assert generated_pem is not None and generated_jwk is not None
    assert not generated_jwk.is_public()


def test_private_key_PEM_JWK_explicit_string():
    generated_pem, generated_jwk = JWT.get_PEM_JWK(mocks.SAMPLE_RSA)

    assert generated_pem is not None and generated_jwk is not None
    assert not generated_jwk.is_public()



import pytest
import okta.models as models
import time
from datetime import datetime
from tests.mocks import MockOktaClient
from okta.errors.okta_api_error import OktaAPIError
from okta.constants import DATETIME_FORMAT


class TestUsersResource:
    """
    Integration Tests for the Users Resource.
    """

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_user(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })

        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Get"
        user_profile.email = "John.Doe-Get@example.com"
        user_profile.login = "John.Doe-Get@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        query_params_create = {"activate": "False"}

        # Create User
        user, resp, err = await test_client.create_user(
            create_user_req, query_params_create)
        assert err is None

        got_user, resp, err = await test_client.get_user(user.id)
        assert got_user.id == user.id

        got_user_with_string, resp, err = await test_client.get_user(
            user_profile.login)
        assert got_user_with_string.id == user.id
        assert user.profile.first_name == user_profile.first_name
        assert user.profile.last_name == user_profile.last_name
        assert user.profile.email == user_profile.email
        assert user.profile.login == user_profile.login

        user_id = user.id

        # Deactivate, then delete created user
        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

        user, _, err = await test_client.get_user(user_id)
        assert user is None
        assert isinstance(err, OktaAPIError)

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_activate_user(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Activate"
        user_profile.email = "John.Doe-Activate@example.com"
        user_profile.login = "John.Doe-Activate@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        # Create query parameters and Create User
        query_params_create = {"activate": "False"}
        user, resp, err = await test_client.create_user(
            create_user_req, query_params_create)
        assert err is None

        # Create query parameters and Activate User
        query_params_activate = {"sendEmail": "False"}
        token, resp, err = await test_client.activate_user(
            user.id, query_params_activate)
        assert err is None
        assert token is not None
        # Ensure correct Object is returned
        assert isinstance(token, models.UserActivationToken)

        # Create query parameters and List Users
        query_params_list = {"filter": "status eq \"ACTIVE\""}
        users, resp, err = await test_client.list_users(query_params_list)
        assert err is None
        # Ensure user is in list
        assert next((usr for usr in users if usr.id ==
                     user.id), None) is not None

        # Deactivate, then delete created user
        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_user_profile(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Update-Profile"
        user_profile.email = "John.Doe-Update-Profile@example.com"
        user_profile.login = "John.Doe-Update-Profile@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        # Create query parameters and Create User
        query_params_create = {"activate": "False"}
        user, _, err = await test_client.create_user(
            create_user_req, query_params_create)
        assert err is None

        # Craft new profile and get user object
        new_profile = user.profile
        NICK_NAME = "JD"
        new_profile.nick_name = NICK_NAME
        updated_user = models.User({"profile": new_profile})

        # Update User with new details
        new_user, _, err = await test_client.update_user(user.id, updated_user)
        assert err is None
        assert new_user is not None

        # Retrieve user and ensure details are updated
        got_user, _, err = await test_client.get_user(user.id)
        assert err is None
        assert got_user.id == user.id
        assert got_user.profile.nick_name == NICK_NAME

        # Deactivate, then delete created user
        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_suspend_user(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Suspend"
        user_profile.email = "John.Doe-Suspend@example.com"
        user_profile.login = "John.Doe-Suspend@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        # Create Query Parameters and Create User
        query_params_create = {"activate": "True"}
        user, _, err = await test_client.create_user(
            create_user_req, query_params_create)
        assert err is None

        # Suspend User
        _, err = await test_client.suspend_user(user.id)
        assert err is None

        # Create query params and List Suspended Users
        query_params_list = {"filter": "status eq \"SUSPENDED\""}
        users, resp, err = await test_client.list_users(query_params_list)
        assert err is None
        # Ensure created user is in list
        assert next((usr for usr in users if usr.id ==
                     user.id), None) is not None

        # Unsuspend User
        _, err = await test_client.unsuspend_user(user.id)
        assert err is None
        # Create query params and List Active Users
        query_params_list = {"filter": "status eq \"ACTIVE\""}
        users, resp, err = await test_client.list_users(query_params_list)
        assert err is None
        # Ensure created user is in list
        assert next((usr for usr in users if usr.id ==
                     user.id), None) is not None

        # Deactivate, then delete created user
        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_change_user_password(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Change-PW"
        user_profile.email = "John.Doe-Change-PW@example.com"
        user_profile.login = "John.Doe-Change-PW@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        # Create Query Parameters and Create User
        query_params_create = {"activate": "True"}
        user, _, err = await test_client.create_user(
            create_user_req, query_params_create)
        assert err is None

        # Give time before changing password
        time.sleep(1)

        # Create new Password
        NEW_PASSWORD = "2ANewPassword!"
        new_password = models.PasswordCredential({
            "value": NEW_PASSWORD
        })

        change_pw_req = models.ChangePasswordRequest({
            "oldPassword": password,
            "newPassword": new_password
        })

        # change password and retrieve user after
        _, _, err = await test_client.change_password(user.id, change_pw_req)
        assert err is None

        got_user, resp, err = await test_client.get_user(user.id)
        assert got_user.id == user.id
        assert err is None
        # Verify password in updated user obj, changed after original
        assert datetime.strptime(got_user.password_changed, DATETIME_FORMAT)\
            > datetime.strptime(user.password_changed, DATETIME_FORMAT)

        # Deactivate, then delete created user
        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_get_reset_password_link_for_user(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Reset-PW-URL"
        user_profile.email = "John.Doe-Reset-PW-URL@example.com"
        user_profile.login = "John.Doe-Reset-PW-URL@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        # Create Query Parameters and Create User
        query_params_create = {"activate": "True"}
        user, _, err = await test_client.create_user(
            create_user_req, query_params_create)

        # Create Query Parameters and get Reset Password Token
        query_params_reset_pw = {"sendEmail": "False"}
        reset_pw_token, _, err = await test_client.reset_password(
            user.id, query_params_reset_pw)
        assert err is None
        assert isinstance(reset_pw_token, models.ResetPasswordToken)
        assert reset_pw_token is not None

        # Deactivate, then delete created user
        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_expire_password_get_temporary(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Expire-PW-URL"
        user_profile.email = "John.Doe-Expire-PW-URL@example.com"
        user_profile.login = "John.Doe-Expire-PW-URL@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        # Create Query Parameters and Create User
        query_params_create = {"activate": "True"}
        user, _, err = await test_client.create_user(
            create_user_req, query_params_create)
        assert err is None

        # Create Query Parameters and get Temporary Password
        temporary_pw, _, err = await\
            test_client.expire_password_and_get_temporary_password(user.id)
        assert err is None
        assert isinstance(temporary_pw, models.TempPassword)
        assert temporary_pw is not None

        # Deactivate, then delete created user
        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_change_recovery_question(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Change-Recovery"
        user_profile.email = "John.Doe-Change-Recovery@example.com"
        user_profile.login = "John.Doe-Change-Recovery@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        # Create Query Parameters and Create User
        query_params_create = {"activate": "True"}
        user, _, err = await test_client.create_user(
            create_user_req, query_params_create)
        assert err is None

        # Create Recovery Question
        NEW_RECOVERY_Q = "What's your favourite security question?"
        NEW_RECOVERY_A = "This One!"

        new_recovery_question = models.RecoveryQuestionCredential({
            "question": NEW_RECOVERY_Q,
            "answer": NEW_RECOVERY_A
        })

        new_user_creds = models.UserCredentials({
            "password": password,
            "recoveryQuestion": new_recovery_question
        })

        # Change Recovery Question
        updated_user_creds, _, err = await\
            test_client.change_recovery_question(user.id, new_user_creds)
        assert err is None
        assert isinstance(updated_user_creds, models.UserCredentials)

        # Create New Password
        NEW_PASSWORD = password.value[::-1]  # reverse string
        new_pw = models.PasswordCredential({
            "value": NEW_PASSWORD
        })
        recovery_answer = models.RecoveryQuestionCredential({
            "answer": NEW_RECOVERY_A
        })

        change_pw_user_creds = models.UserCredentials({
            "password": new_pw,
            "recoveryQuestion": recovery_answer
        })

        # Create New Password using Recovery Question answer
        _, _, err = await test_client.forgot_password_set_new_password(
            user.id, change_pw_user_creds)
        assert err is None

        # Get User and verify password was changed
        got_user, _, err = await test_client.get_user(user.id)
        assert got_user.id == user.id
        assert err is None
        DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
        assert datetime.strptime(got_user.password_changed, DATETIME_FORMAT)\
            > datetime.strptime(user.password_changed, DATETIME_FORMAT)

        # Deactivate, then delete created user
        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_assign_user_to_role(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Assign-User-Role"
        user_profile.email = "John.Doe-Assign-User-Role@example.com"
        user_profile.login = "John.Doe-Assign-User-Role@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        # Create Query Parameters and Create User
        query_params_create = {"activate": "True"}
        user, _, err = await test_client.create_user(
            create_user_req, query_params_create)
        assert err is None

        # Create Assign Role Request with Roletype Enum
        USER_ADMIN = models.RoleType.USER_ADMIN
        assign_role_req = models.AssignRoleRequest({
            "type": USER_ADMIN
        })

        # Assign Role to User
        _, _, err = await test_client.assign_role_to_user(user.id,
                                                          assign_role_req)
        assert err is None

        # Get Roles for user and ensure role assigned is found
        roles, _, err = await test_client.list_assigned_roles_for_user(user.id)
        found_role = next((role for role in roles if role.type ==
                           USER_ADMIN), None)
        assert found_role is not None

        # Remove assigned role from user
        _, err = await test_client.remove_role_from_user(user.id,
                                                         found_role.id)
        assert err is None

        # Get Roles for user and ensure role assigned is NOT found
        roles, _, err = await test_client.list_assigned_roles_for_user(
            user.id)
        found_role = next((role for role in roles if role.type ==
                           USER_ADMIN), None)
        assert found_role is None

        # Deactivate, then delete created user
        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_user_group_target_to_role(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Group-Target-Role-Assign"
        user_profile.email = "John.Doe-Group-Target-Role-Assign@example.com"
        user_profile.login = "John.Doe-Group-Target-Role-Assign@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        # Create Query Parameters and Create User
        query_params_create = {"activate": "True"}
        user, _, err = await test_client.create_user(
            create_user_req, query_params_create)
        assert err is None

        # Create Group Object
        NEW_GROUP_NAME = "Group-Target-Test-Assign"
        new_group_profile = models.GroupProfile({
            "name": NEW_GROUP_NAME
        })
        new_group = models.Group({
            "profile": new_group_profile
        })

        # Create Group
        group, _, err = await test_client.create_group(new_group)
        assert err is None

        # Create request to assign role to user
        USER_ADMIN = models.RoleType.USER_ADMIN
        assign_role_req = models.AssignRoleRequest({
            "type": USER_ADMIN
        })

        # Assign Role to User
        user_role, _, err = await test_client.assign_role_to_user(
            user.id, assign_role_req)
        assert err is None

        # Add Group Target to the Role
        _, err = await test_client.add_group_target_to_role(
            user.id, user_role.id, group.id)
        assert err is None

        # Retrieve group targets for role and ensure added one is there
        groups, _, err = await test_client.list_group_targets_for_role(
            user.id, user_role.id)
        assert next((grp for grp in groups if grp.id ==
                     group.id), None) is not None

        # Create another group to add
        NEW_GROUP_NAME = "Temp-Group-Target-Test-Assign"
        new_group_profile = models.GroupProfile({
            "name": NEW_GROUP_NAME
        })
        new_group = models.Group({
            "profile": new_group_profile
        })

        # Create 2nd group
        temp_group, _, err = await test_client.create_group(new_group)
        assert err is None

        # Add new group target to role and remove original
        _, err = await test_client.add_group_target_to_role(
            user.id, user_role.id, temp_group.id)
        assert err is None
        _, err = await test_client.remove_group_target_from_role(
            user.id, user_role.id, group.id)
        assert err is None

        # Deactivate, then delete created user
        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

        _, err = await test_client.deactivate_or_delete_user(user.id)
        assert err is None

        # Delete groups created
        await test_client.delete_group(group.id)
        await test_client.delete_group(temp_group.id)

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_user_pagination(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create 2 user profiles
        first_user_profile = models.UserProfile()
        first_user_profile.first_name = "John"
        first_user_profile.last_name = "Doe-Paginate-1"
        first_user_profile.email = "John.Doe-Paginate-1@example.com"
        first_user_profile.login = "John.Doe-Paginate-1@example.com"

        second_user_profile = models.UserProfile()
        second_user_profile.first_name = "John"
        second_user_profile.last_name = "Doe-Paginate-2"
        second_user_profile.email = "John.Doe-Paginate-2@example.com"
        second_user_profile.login = "John.Doe-Paginate-2@example.com"

        first_create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": first_user_profile
        })
        second_create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": second_user_profile
        })

        # Create Query Parameters and Create 2 Users
        query_params_create = {"activate": "False"}
        user1, _, err = await test_client.create_user(
            first_create_user_req, query_params_create)
        assert err is None
        user2, _, err = await test_client.create_user(
            second_create_user_req, query_params_create)
        assert err is None

        # Create Query Parameters and List Users
        query_params_limit = {"limit": "1"}  # Retrieve one user
        usr_list, resp, err = await test_client.list_users(query_params_limit)
        assert err is None
        assert len(usr_list) == 1
        assert isinstance(usr_list[0], models.User)
        assert resp.has_next()  # ensure next can be determined

        # Retrieve next page of results
        second_usr_list, err = await resp.next()
        assert err is None
        assert len(second_usr_list) == 1
        assert isinstance(second_usr_list[0], models.User)

        # Deactivate & delete both users
        _, err = await test_client.deactivate_or_delete_user(user1.id)
        assert err is None
        _, err = await test_client.deactivate_or_delete_user(user1.id)
        assert err is None

        _, err = await test_client.deactivate_or_delete_user(user2.id)
        assert err is None
        _, err = await test_client.deactivate_or_delete_user(user2.id)
        assert err is None

        # Ensure deletion
        found_user, _, err = await test_client.get_user(user1.id)
        assert found_user is None
        assert isinstance(err, OktaAPIError)
